/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "oirSupport.h"
#include <string.h>
#define TINY 1.e-20

void minrtsLudCmp(float *a, int n, int *indx, float *d) {

   int i, imax,j,k;
   float big,dum,sum,temp;
   float *vv;

   vv = (float *)malloc(n*sizeof(float));

   *d=1.;

   for (i=0;i<n;i++) {
      big = 0.0;
      for (j=0;j<n;j++) {
         temp = fabs(a[n*i+j]);
         if (temp > big) big = temp;
      }
      vv[i] = 1./big;
   }

   for (j=0; j<n;j++){
      for (i=0;i<j;i++) {
         sum = a[n*i+j];
         for (k=0;k<i;k++) sum -= a[i*n+k]*a[k*n+j];
         a[i*n+j] = sum;
      }
      big = 0.0;
      for (i=j;i<n;i++) {
         sum = a[n*i+j];
         for (k=0;k<j;k++) sum -= a[n*i+k] * a[n*k+j];
         a[n*i+j] = sum;
         dum = vv[i]*fabs(sum);
         if (dum >=big) {
            big = dum;
            imax = i;
         }
      }
      if (j != imax) {
         for (k=0;k<n;k++) {
            dum=a[n*imax+k];
            a[n*imax+k] = a[n*j+k];
            a[n*j+k] = dum;
         }
         *d = -(*d);
         vv[imax] = vv[j];
      }indx[j] = imax;
      if (a[n*j+j] == 0.0) a[n*j+j] = TINY;

      if (j != (n-1)) {
         dum = 1./(a[j*n+j]);
         for (i=j+1;i<n;i++) a[n*i+j] *= dum;
      }
   }
   free(vv);
   return;
}

void minrtsLubKsb(float *a, int n, int *indx, float *b) {
   int i,ii,ip,j;
   float sum;

   ii = -1;
   for (i=0;i<n;i++) {
      ip = indx[i];
      sum = b[ip];
      b[ip] = b[i];
      if (ii >= 0)
         for (j=ii;j<=i-1;j++) sum -= a[n*i+j]*b[j];
      else if (sum) ii=i;
      b[i] = sum;
   }
   for (i=n-1;i>=0;i--) {
      sum = b[i];
      for (j=i+1;j<n;j++) sum -=a[n*i+j]*b[j];
      b[i] = sum/a[n*i+i];
   }
   return;
}

/***************************************************************
 * Fit a polynomial to a set of data, given weights
 * INPUTS:
 *    int      order     order of polynomial; 0 = constant, 1=linear etc.
 *    int      nData     number of data, coordinate, and weight points
 *    float*   xCoord    independent variable at each measured point
 *    float*   data      dependent variable at each measured point
 *    float*   weight    relative weight to apply to each point
 * OUTPUTS:
 *    float*   coeff     preallocated [1+order] float array to contain answers
 **************************************************************** */
void minrtsPolyFit(int order, int nData, float *xCoord, float *data, 
   float *weight, float *coeff) {

   float *matrix, *vector;
   float d, pj, pk; 
   int *indx;
   int i,j,k;

   matrix = (float *)malloc((1+order)*(1+order)*sizeof(float));
   vector = (float *)malloc((1+order)*sizeof(float));
   indx = (int *)malloc((1+order)*sizeof(float));
   memset(matrix, 0, (1+order)*(1+order)*sizeof(float));
   memset(vector, 0, (1+order)*sizeof(float));

   for (i=0; i<nData; i++) {
      for (j=0;j<=order; j++) {
         if (j == 0) pj = 1; else pj = pow(xCoord[i], j);
         vector[j] += data[i]*weight[i]*pj;
         for (k=0;k<=order; k++) {
            if (k == 0) pk = 1; else pk = pow(xCoord[i], k);
            matrix[k+(1+order)*j] += weight[i] * pj * pk;
         }
      }
   }
   minrtsLudCmp(matrix, order+1, indx, &d);
   minrtsLubKsb(matrix, order+1, indx, vector);
   for (i=0;i<=order;i++) coeff[i] = vector[i];

   free(matrix);
   free(vector);
   free(indx);
   return;
   } /* end of minrtsPolyFit */
/***************************************************************
 * Expand a polynomial given coefficients and an array of input coordinates
 * INPUTS:
 *    int      order     order of polynomial; 0 = constant, 1=linear etc.
 *    int      nData     number of coordinate
 *    float*   xCoord    independent variable at each measured point
 *    float*   coeff     [1+order] coefficients; lowest order (constant) first
 * OUTPUTS:
 *    float*   data      preallocated [nData] array to receive answers
*************************************************************************/
void minrtsPolyExpand (int order, int nData, float* xCoord, float* coeff, float* data) {

   int idata,iorder;
   float x,y;

   for (idata=0;idata<nData;idata++){
      x  = xCoord[idata];
      y  = 0.;
      for (iorder=order;iorder>=0;iorder--) y = x*y + coeff[iorder];
      data[idata] = y;
   }
   return;
   } /* end of minrtsPolyExpand

/***************************************************************
 * Split a blank-separated set of file names into separate character strings
 * INPUTS:
 *    char *string     blank-separated list
 * OUTPUTS:
 *    int    nString      number of strings found
 * RETURNS
 *    char **   array of output file names; these have been malloced and
 *              should be freed by calling program killFileList
 **************************************************************** */
char** oirStrSplit (char *string, int *nString) {
   char *blank = " ";
   char *oldString;
   int   nS, iS ;
   char *tok;
   char *stringCopy;
   char **output;

   *nString = 0;
/*            copy input string                */
   stringCopy = (char *)malloc(sizeof(char) * (1+strlen(string)));
   strcpy (stringCopy, string);
   oldString = stringCopy;
/*          find pointer to first blank delimited string   */
   tok = strtok(oldString, blank);
   if (tok == NULL) return NULL;
   oldString = NULL;
/*          get pointers to the rest of the sub strings */
   for (nS = 0; tok != NULL; nS++) {
      tok = strtok(oldString, blank);
   }
   *nString = nS;
/*          malloc output arrays and copy strings */
   output = (char **)malloc(nS * sizeof(char *));
   strcpy (stringCopy, string);
   oldString = stringCopy;
   blank = " ";
   for (iS=0; iS<nS; iS++) {
      tok = strtok(oldString, blank);
      oldString = NULL;
      output[iS] = (char *)malloc(sizeof(char) * (1+strlen(tok)));
      strcpy(output[iS], tok);
   }
   free(oldString);
   return output;
} /* end of oirStrSplit */ 

/***************************************************************
 * kill space associated with array of file names
 * INPUTS:
 *    char **string     array of file names
 *    int    nFiles     number of file names
 **************************************************************** */
void oirKillFileList(char **fileList, int nFiles) {
   int iFile;
   for (iFile=0; iFile<nFiles;iFile++) {free(fileList[iFile]);}
   free(fileList);
   return;
} /* end of oirKillFileList */

void oirGetMaskImages(char *maskFileName, float *maskImages[], 
   int nRegion, int nxList[], int nyList[], int *ierr){

   fitsfile *maskFile;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_table_def *idata;
   imaging_fdata_row *inData, *inData1;

   char maskType[72];
   int image;
   int i, ix, iy, npix, nx;
   int iRegion;
   float x, sigma;

   fits_open_file (&maskFile, maskFileName, READONLY, ierr);
   if (*ierr != 0) {
      printf("fits openMaskfile of %s gave error %i",maskFileName, *ierr);
      printf("%s \n",strerror(errno));
      return;
   }
   i = fits_read_key(maskFile, TSTRING, "MASKTYPE", maskType, 0, ierr);
   if (*ierr != 0) image = 1; else {
      if(strstr(maskType, "IMAGE") != 0) image = 1;
         else image = 0;
      }
   *ierr = 0;
   DetectorDef = read_imaging_detector_table(maskFile, ierr);
   if (*ierr !=0) {
      printf("error %i opening detector table in %s\n",
         *ierr, maskFileName);
      printf("%s \n",strerror(errno));
      return ;
   }

   idata = open_imaging_fdata_table(maskFile, ierr);
   if (*ierr !=0) {
      printf("error %i opening imaging data in file %s\n",
         *ierr, maskFileName);
      printf("%s \n",strerror(errno));
      return;
   }
   inData = make_imaging_fdata_row(idata);
   if (image ==1 ) {   /* mask is an image */
      i = read_imaging_fdata_table(maskFile, idata, inData, 1, ierr);
/*
           check if dimensions agree
*/
      for (iRegion=0; iRegion<nRegion;iRegion++) {
         if(DetectorDef->naxis[iRegion][0] != nxList[iRegion] ||
            (DetectorDef->naxis[iRegion][1] != nyList[iRegion]) ) {
            printf("mask dimensions of region %i dont agree with data dimensions\n",
               iRegion);
            *ierr = 1;
            return;
         }
/*
           allocate array and copy data
*/
         npix = nxList[iRegion]*nyList[iRegion];
         maskImages[iRegion] = (float *)malloc(npix*sizeof(float));
         for (i=0;i<npix;i++) maskImages[iRegion][i] = inData->data[iRegion][i];
      }
   } else {   /* mask is a bunch of gaussian components */
      inData1 = make_imaging_fdata_row(idata);
      i = read_imaging_fdata_table(maskFile, idata, inData,  1, ierr); /* AData */
      i = read_imaging_fdata_table(maskFile, idata, inData1, 2, ierr); /* BData */
/*
           check if dimensions agree
*/
      for (iRegion=0; iRegion<nRegion;iRegion++) {
         if(DetectorDef->naxis[iRegion][0] != nxList[iRegion]) {
            printf("mask dimensions of region %i dont agree with data dimensions\n",
               iRegion);
            *ierr = 1;
            return;
         }
/*
           allocate array 
*/
         nx = nxList[iRegion];
         npix = nx*nyList[iRegion];
         maskImages[iRegion] = (float *)malloc(npix*sizeof(float));
/*
           average A & B and compute gaussian components
*/
         for (ix=0;ix<nx;ix++) {
            for (iy=0;iy<3;iy++) inData->data[iRegion][ix+nx*iy] =
               0.5*(inData->data[iRegion][ix+nx*iy]+inData1->data[iRegion][ix+nx*iy]);

            for (iy=0;iy<nyList[iRegion];iy++) {
               x = inData->data[iRegion][ix] - (iy+1);
               sigma = inData->data[iRegion][ix+2*nx];
               x = 0.5*x*x/(sigma*sigma);
               if (x>10) x=10.;
               maskImages[iRegion][ix + nx*iy] = inData->data[iRegion][ix+nx]*exp(-x);
            } /* end of iy loop */
         }
      }
      kill_imaging_fdata_row(inData1);
   } /* end of image/gauss if else */
    i= close_imaging_fdata_table(maskFile, idata, ierr);
    kill_imaging_fdata_row(inData);
   return;
} /* end of oirGetMaskImages */

void midiRectXWindow(minrtsFrameData *data, 
   imaging_detector_table_def *coord,
   int refWindow, int *ierr) {

   int iregion, nregion;
   int ix, nx;
   int iy, ny;

   float *coordData, *refCoordData;
   float *outWork;
   int   *refIndex;
   int iCoord, jCoord, iRef, nRef, nCoord;
   int i,j,k;
   float x;
   int n2;
   int dec;

   nregion = data->nregion;
/*
      loop over regions, check compatibility
      and get coordinate data
*/
   if (nregion != coord->nregion) {
      printf("data nregions = %i; coord regions =%i\n", nregion, coord->nregion);
      *ierr=2;
      return;
   }
   if (nregion < refWindow) {
      printf("refWindow %i is greater than nregion %i\n", refWindow, nregion);
      *ierr=1;
      return;
   }
   for (iregion=0; iregion<nregion; iregion++) {
      nx = data->nx[iregion];
      ny = data->ny[iregion];
      if ((nx != coord->naxis[iregion][0]) || (ny != coord->naxis[iregion][1])) {
         printf("naxis for window %i %i %i NE coord values %i %i\n",
            iregion+1, nx, ny, coord->naxis[iregion][0], coord->naxis[iregion][1]);
         *ierr = 2;
         return;
      }
   }  /* end of iregion loop */
   refCoordData = imaging_detector_wcs(coord, refWindow);
/*
      find monotonic region of refCoord and prepare for
      spline routine (which needs monotonically increasing
      abscissa
*/
   nx = data->nx[refWindow-1];
   n2 = (int) nx/2;
   refIndex = (int *)malloc(nx*sizeof(int));
   for (i=0;i<nx;i++) refIndex[i] = i;
   dec = refCoordData[n2] < refCoordData[n2-1]; /* decreasing values */
   iRef = n2;
   if (dec) {
      while ((iRef >= 0) && (refCoordData[iRef] < refCoordData[iRef-1]))
         --iRef;
      jCoord = n2;
      while ((jCoord < nx-1) && (refCoordData[jCoord] < refCoordData[jCoord-1]))
         ++jCoord;
      nRef = jCoord - iRef + 1;
/*   
     flip both coordinates and indices
*/
      j = jCoord;
      for (i=iRef;i<iRef+nRef/2;i++) {
         x = refCoordData[i];
         refCoordData[i] = refCoordData[j];
         refCoordData[j] = x;
         k = refIndex[i];
         refIndex[i] = refIndex[j];
         refIndex[j] = k;
         --j;
      }
   } else {  /* increasing values */
      while ((iRef >= 0) && (refCoordData[iRef] > refCoordData[iRef-1]))
         --iRef;
      jCoord = n2;
      while ((jCoord < nx-1) && (refCoordData[jCoord] > refCoordData[jCoord-1]))
         ++jCoord;
      nRef = jCoord - iRef + 1;
   }
/*     
   now loop over regions 
   flip coordinates if necessary to get increasing values
   and interpolate rows to reference coord
*/
   for (iregion=0; iregion<nregion; iregion++) {
      coordData = imaging_detector_wcs(coord, 1+iregion);
      nx = data->nx[iregion];
      ny = data->ny[iregion];
/*     condition input coordinates if necessary */
    
      n2 = (int) nx/2;
      dec = coordData[n2] < coordData[n2-1]; /* decreasing values */
      iCoord = n2;
      if (dec) {
         while ((iCoord >= 0) && (coordData[iCoord] < coordData[iCoord-1]))
            --iCoord;
         jCoord = n2;
         while ((jCoord < nx-1) && (coordData[jCoord] < coordData[jCoord-1]))
         ++jCoord;
         nCoord = jCoord - iCoord + 1;
/*   
     flip input coordinates 
*/
         j = jCoord;
         for (i=iCoord;i<iCoord+nCoord/2;i++) {
            x = coordData[i];
            coordData[i] = coordData[j];
            coordData[j] = x;
            --j;
         }
      } else {  /* increasing values */
         while ((iCoord >= 0) && (coordData[iCoord] > coordData[iCoord-1]))
            --iCoord;
         jCoord = n2;
         while ((jCoord < nx-1) && (coordData[jCoord] > coordData[jCoord-1]))
            ++jCoord;
         nCoord = jCoord - iCoord + 1;
      }
/*
         loop over rows, flip data if necessary interpolate
         and flip interpolated data (if necessary)
*/
      for (iy=0;iy<ny;iy++) {
         if (dec) {
            j = jCoord + iy*nx;
            for (i=iCoord+iy*nx;i<iCoord+iy*nx+nCoord/2;i++) {
               x = data->data[iregion][i];
               data->data[iregion][i] = data->data[iregion][j];
               data->data[iregion][j] = x;
               --j;
            }
         }
         outWork = minrtsSpline(nCoord, coordData+iCoord, 
            data->data[iregion]+iy*nx+iCoord, nRef, refCoordData+iRef, 0.1);
         for (ix=0;ix<nx;ix++) data->data[iregion][ix + iy*nx] = 0.0;
         for (i=0;i<nRef;i++) data->data[iregion][iy*nx+refIndex[iRef+i]] =
            outWork[i];
         free (outWork);
      } /* loop over y */
      free (coordData);
   }  /* end of iregion interpolation loop */
   free(refCoordData);
   free(refIndex);
   return;
}

void midiRectYWindow(minrtsFrameData *data, 
   imaging_detector_table_def *coord, int forward, int *ierr) {

   int iregion, nregion;
   int ix, nx;
   int iy, ny;

   float *coordData;
   float *oldY, *newY;
   float *inWork, *outWork;

   nregion = data->nregion;
/*
      loop over regions, check compatibility
      and get coordinate data
*/
   if (nregion != coord->nregion) {
      printf("data nregions = %i; coord regions =%i\n", nregion, coord->nregion);
      *ierr=2;
      return;
   }
   for (iregion=0; iregion<nregion; iregion++) {
      nx = data->nx[iregion];
      ny = data->ny[iregion];
      if ((nx != coord->naxis[iregion][0]) || (ny != coord->naxis[iregion][1])) {
         printf("naxis for window %i %i %i NE coord values %i %i\n",
            iregion+1, nx, ny, coord->naxis[iregion][0], coord->naxis[iregion][1]);
         *ierr = 2;
         return;
      }
   }  /* end of iregion check loop */
/*     
   now loop over regions and slide rows up and down 
*/
   for (iregion=0; iregion<nregion; iregion++) {
      coordData = imaging_detector_wcs(coord, 1+iregion);
      nx = data->nx[iregion];
      ny = data->ny[iregion];
      oldY = (float *)malloc(ny * sizeof(float));
      newY = (float *)malloc(ny * sizeof(float));
      inWork= (float *)malloc(ny * sizeof(float));
      for (iy=0;iy<ny;iy++) newY[iy] = (float)(1+iy-coord->crpix[iregion][1]);
      for (ix=0; ix<nx; ix++) {
         for (iy=0;iy<ny;iy++) {
            inWork[iy] = data->data[iregion][ix + iy*nx];
            oldY[iy] = coordData[ix + nx*(ny+iy)];
         }
         if (forward)outWork = minrtsSpline(ny, oldY, inWork, ny, newY, 0.1);
         else outWork = minrtsSpline(ny, newY, inWork, ny, oldY, 0.1);
         for (iy=0;iy<ny;iy++) data->data[iregion][ix + iy*nx] = outWork[iy];
         free (outWork);
      } /* loop over x */
      free (oldY);
      free (newY);
      free (inWork);
      free (coordData);
   }  /* end of iregion slide loop */


   return;
}

void midiRectXYWindow(minrtsFrameData *data, 
   imaging_detector_table_def *coord,
   int refWindow, int *ierr) {

   midiRectXWindow(data, coord, refWindow, ierr);
   if ((*ierr) == 0) midiRectYWindow(data, coord, 1, ierr);
   return;
}


void midiDiffRectYWindow(minrtsFrameData *data, 
   imaging_detector_table_def *coord, 
   int refWindow, int *ierr) {

   int iregion, nregion;
   int ix, nx;
   int iy, ny;

   float *coordData, *refCoordData;
   float *oldY, *newY;
   float *inWork, *outWork;

   nregion = data->nregion;
/*
      loop over regions, check compatibility
      and get coordinate data
*/
   if (nregion != coord->nregion) {
      printf("data nregions = %i; coord regions =%i\n", nregion, coord->nregion);
      *ierr=2;
      return;
   }
   for (iregion=0; iregion<nregion; iregion++) {
      nx = data->nx[iregion];
      ny = data->ny[iregion];
      if ((nx != coord->naxis[iregion][0]) || (ny != coord->naxis[iregion][1])) {
         printf("naxis for window %i %i %i NE coord values %i %i\n",
            iregion+1, nx, ny, coord->naxis[iregion][0], coord->naxis[iregion][1]);
         *ierr = 2;
         return;
      }
   }  /* end of iregion check loop */
/*     
   now loop over regions and slide rows up and down 
*/
   refCoordData = imaging_detector_wcs(coord, refWindow);
   for (iregion=0; iregion<nregion; iregion++) {
      coordData = imaging_detector_wcs(coord, 1+iregion);
      nx = data->nx[iregion];
      ny = data->ny[iregion];
      oldY = (float *)malloc(ny * sizeof(float));
      newY = (float *)malloc(ny * sizeof(float));
      inWork= (float *)malloc(ny * sizeof(float));
         
      for (ix=0; ix<nx; ix++) {
         for (iy=0;iy<ny;iy++) {
            inWork[iy] = data->data[iregion][ix + iy*nx];
            oldY[iy]   = coordData[ix + nx*(ny+iy)];
            newY[iy]   = refCoordData[ix + nx*(ny+iy)];
         }
/*         outWork = minrtsSpline(ny, newY, inWork, ny, oldY, 0.1); */
         outWork = minrtsSpline(ny, oldY, inWork, ny, newY, 0.1); 
         for (iy=0;iy<ny;iy++) data->data[iregion][ix + iy*nx] = outWork[iy];
         free (outWork);
      } /* loop over x */
      free (oldY);
      free (newY);
      free (inWork);
      free (coordData);
   }  /* end of iregion slide loop */


   return;
}

void midiDiffRectXYWindow(minrtsFrameData *data, 
   imaging_detector_table_def *coord,
   int refWindow, int *ierr) {

   midiRectXWindow(data, coord, refWindow, ierr); 
   if ((*ierr) == 0) midiDiffRectYWindow(data, coord, refWindow, ierr);
   return;
}
/***************************************************************
* get the last part of the file name only, excluding the directory path
* INPUTS:
*    char *path     full path to file
**************************************************************** */

char *oirFileRoot(char *path) {
   char *slash;
   slash = strrchr(path, '/');
   if (slash == NULL) return path;
   return 1+slash;
}

/*********************************************************
 * Two dimensional brute force cross correlation of two
 * real arrays
 * INPUTS:
 *    int      nx     number of data points in x-direction
 *    int      ny     number of data points in y-direction
 *    int      dxmin  max leftward  shift of array2 wrt array1
 *    int      dxmax  max rightward shift of array2 wrt array1
 *    int      dymin  max downward  shift of array2 wrt array1
 *    int      dymax  max upward    shift of array2 wrt array1
 *    float   *array1 first array
 *    float   *array2 2nd array, may be same as first ->autocorrelation
 * OUTPUTS:    float* array of (dxmax-dxmin+1)*(dymax-dymin+1) correlations
 *             allocated here.  Must be free(ed) by user
*********************************************************/
   float * minrtsBFRCross(int nx, int ny, int dxmin, int dxmax, 
      int dymin, int dymax, float *array1, float *array2) {

   int xout, yout, nout;
   int xfirst1, xlast1, xfirst2;
   int yfirst1, ylast1, yfirst2;
   int ix, iy, jx, jy;
   float *a1, *a2;
   float *cc;
   float sum;

   xout = dxmax-dxmin+1;
   yout = dymax-dymin+1;
   nout = xout * yout;

   cc = (float *)malloc(nout * sizeof(float));

   for (iy = dymin; iy<=dymax; iy++) {
      if (iy>=0) {
         yfirst1 = iy;
         ylast1  = ny;
         yfirst2 = 0;
      } else {
         yfirst1 = 0;
         ylast1  = ny+iy;
         yfirst2 = -iy;
      }
      for (ix=dxmin; ix<=dxmax;ix++){
         if (ix>=0) {
            xfirst1 = ix;
            xlast1  = nx;
            xfirst2 = 0;
         } else {
            xfirst1 = 0;
            xlast1  = nx+ix;
            xfirst2 = -ix;
         }
         
         sum   = 0.;
         for (jy=yfirst1;jy<ylast1;jy++) {
            a1 = array1 + nx*jy + xfirst1;
            a2 = array2 + nx*(jy+yfirst2-yfirst1)+xfirst2;
            for (jx=xfirst1;jx<xlast1;jx++) sum += (*a1++)*(*a2++);
         }
         cc[ix+dxmin + (iy+dymin)* xout] = 
            sum/((ylast1-yfirst1)*(xlast1-xfirst1));
      }
   }
   return cc;
}   /* end of minrtsBFRCross */

/*********************************************************
 * Two dimensional brute force cross correlation of two
 * complex arrays stored as "pseudocomplex" 
 *       output(i,j) = <array1*conj(shift(array2), i,j)>
 *       where the ranges of i,j are specified by dxmin,dymin,dxmax,dymax
 * INPUTS:
 *    int      nx     number of data points in x-direction
 *    int      ny     number of data points in y-direction
 *    int      dxmin  max leftward  shift of array2 wrt array1
 *    int      dxmax  max rightward shift of array2 wrt array1
 *    int      dymin  max downward  shift of array2 wrt array1
 *    int      dymax  max upward    shift of array2 wrt array1
 *    float   *array1 first array
 *    float   *array2 2nd array, may be same as first ->autocorrelation
 * OUTPUTS:    float* array of (dxmax-dxmin+1)*(dymax-dymin+1) correlations
 *             allocated here.  Must be free(ed) by user
*********************************************************/
   float * minrtsBFCCross(int nx, int ny, int dxmin, int dxmax, 
      int dymin, int dymax, float *array1, float *array2) {

   int xout, yout, nout;
   int xfirst1, xlast1, xfirst2;
   int yfirst1, ylast1, yfirst2;
   int ix, iy, jx, jy;
   float *a1, *a2;
   float *cc;
   float sumR, sumI;

   xout = dxmax-dxmin+1;
   yout = dymax-dymin+1;
   nout = xout * yout;

   cc = (float *)malloc(2*nout * sizeof(float));

   for (iy = dymin; iy<=dymax; iy++) {
      if (iy>=0) {
         yfirst1 = iy;
         ylast1  = ny;
         yfirst2 = 0;
      } else {
         yfirst1 = 0;
         ylast1  = ny+iy;
         yfirst2 = -iy;
      }
      for (ix=dxmin; ix<=dxmax;ix++){
         if (ix>=0) {
            xfirst1 = ix;
            xlast1  = nx;
            xfirst2 = 0;
         } else {
            xfirst1 = 0;
            xlast1  = nx+ix;
            xfirst2 = -ix;
         }
         
         sumR   = 0.;
         sumI   = 0.;
         for (jy=yfirst1;jy<ylast1;jy++) {
            a1 = array1 + 2*(nx*jy + xfirst1);
            a2 = array2 + 2*(nx*(jy+yfirst2-yfirst1)+xfirst2);
            for (jx=xfirst1;jx<xlast1;jx++) {
               sumR +=  (*a1)*(*a2)     + (*(a1+1))*(*(a2+1));
               sumI += -(*a1)*(*(a2+1)) + (*(a1+1))*(*a2);
               a1 = a1+2;
               a2 = a2+2;
            }
         }
         cc[2*(ix+dxmin + (iy+dymin)* xout)] = 
            sumR/((ylast1-yfirst1)*(xlast1-xfirst1));
         cc[2*(ix+dxmin + (iy+dymin)* xout)+1] = 
            sumI/((ylast1-yfirst1)*(xlast1-xfirst1));
      }
   }
   return cc;
}   /* end of minrtsBFCCross */

